import copy

# Safe import of OrderedDict
try:
    from collections import OrderedDict
except ImportError:
    from ordered_dict_26 import OrderedDict


class DataStructure(OrderedDict):
    """ Mutable structure, similar to OrderedDicts, that stores data read from files

    This is the class returned by read_pds4(), read_table() and read_array(). If containing multiple File Areas, or
    multiple data classes (e.g. a Table_Binary and an Array_2D_Image), or table groups then DataStructure will have
    nested DataStructures inside it for each one of these. Effectively this class is a (potentially nested) OrderedDict,
    and can be treated/accessed/used as one in almost all ways. An important exception is that assignments
    such as dict[key] = value are done via dict.set(key, value, meta_data). The meta_data stores information about the
    value (e.g. if value is a table field, then meta_data would be a Field object containing things like field_format).
    Numeric keys may also be used to access values in this object, similar to a List.
    """

    def __init__(self):
        self._meta_data = {}
        super(DataStructure, self).__init__()

    def __getitem__(self, key):

        if isinstance(key, (int, long)):
            item = self[self.keys()[key]]
        else:
            item = super(DataStructure, self).__getitem__(key)

        return item

    def __setitem__(self, key, value, dict_setitem=dict.__setitem__):
        raise TypeError('DataStructure object does not support item assignment, use .set() method')

    # This method overwrites copy.deepcopy(DataStructure). Note that for anything but the smallest data this method
    # will be extremely expensive unless use_numpy is on
    def __deepcopy__(self, memo):

        deepcopied_data_structure = DataStructure()

        # Recursively deepcopy all values
        for key, value in self.iteritems():
            deepcopied_data_structure.set(key, copy.deepcopy(value, memo), copy.deepcopy(self.meta_data(key), memo))

        return deepcopied_data_structure

    def set(self, key, data, meta_data):

        if isinstance(key, (int, long)):
            raise NameError('DataStructure object does not support integer keys')

        self._meta_data[key] = meta_data
        super(DataStructure, self).__setitem__(key, data)

    def meta_data(self, key):

        if isinstance(key, (int, long)):
            meta_data = self._meta_data[self._meta_data.keys()[key]]
        else:
            meta_data = self._meta_data[key]

        return meta_data

    # Returns an equivalent, uncopied, list (stripping the meta_data)
    def to_list(self):

        list_dt = list()

        # Recursively convert all nested DataStructures to List
        for key, value in self.iteritems():

            if isinstance(value, DataStructure):
                list_dt.append(value.to_list())
            else:
                list_dt.append(value)

        return list_dt

    # Returns an equivalent, uncopied, OrderedDict (stripping the meta_data)
    def to_ordered_dict(self):

        ordered_dict = OrderedDict(self)

        # Recursively convert all nested DataStructures to OrderedDicts
        for key, value in ordered_dict.iteritems():

            if isinstance(value, DataStructure):
                ordered_dict[key] = value.to_ordered_dict()

        return ordered_dict


class DataObject(object):
    """ Stores meta data about a single Data Object (e.g. Table_Character,  Array_2D_Image,  Array_3D, etc) """

    def __init__(self, name, type, parent_filename):

        self.name = name
        self.type = type
        self.parent_filename = parent_filename

