import xml.etree.ElementTree as ET
from xml.parsers.expat import ExpatError
from os import path

import pds4_read_table
from pds4_general_objects import DataStructure
from pds4_logging import logger_init

# Initialize the logger
logger = logger_init()

#################################


def pds4_read(filename, use_numpy=False, quiet=False, object_num=None, object_name=None, object_lid=None):
    """
        Reads PDS4 compliant data into Python.

        NOTES:

            Currently supports only the objects Table_Character and Table_Binary. Packed bit fields in Table_Binary are
            not yet supported.

        Parameters:

            filename: str
                The filename, including full or relative path if necessary, of the PDS4 label describing the data
            quiet: bool
                Suppresses all info/warnings from being output
            use_numpy: bool
                Returned data will be a NumPy array and use NumPy data types
            object_num: str
                If given then only reads the data object specified by zero-based order (integer)
            object_name: str
                If given then only reads the data object specified by name
            object_lid: str
                If given then only reads the data object specified by local identifier

        Returns:

            data: DataStructure
                Read-in data and meta-data. Can be treated/accessed/used like an OrderedDict.

                Use to_ordered_dict() method to retrieve as OrderedDict
                Use to_list() method to retrieve as List

        Example usage:

            data = pds4_read('Example_Label.xml')

            Example Label:

                Table: unnamed
                    Field: order
                    Field: wavelength - (1)
                Table: Observations
                    Field: order
                    Group: unnamed; repetitions: 2
                        Field: pos_vector - (2)

            Accessing Example Label data:

                You may use any combination of Dict or List like indicies.

                (1) data['TABLE_0']['wavelength']
                (1) data['TABLE_0'][1]
                (1) data[0]['Wavelength']
                (1) data[0][1]

                (2) data['Observations']['GROUP_0']['pos_vector'][0]
                    data['Observations']['GROUP_0']['pos_vector'][1]

                (2) data['Observations']['GROUP_0'][0][0]
                    data['Observations'][1]['pos_vector'][1]

                (2) data[0][1][0]
                    data[0][1][1]

            Accessing Example Label meta data:

                As with accessing data, you may use any combination of Dict or List like indicies.

                (1) data['TABLE_0'].meta_data('wavelength').unit
                (1) data[0].meta_data(1).unit

                (2) data['Observations']['GROUP_0'].meta_data('pos_vector')[0].unit
                (2) data[0][1].meta_data(0)[1].unit

    """

    # Set logger to only emit error messages if requested
    if quiet:
        logger.quiet()

    # Parse XML, strip all namespaces
    try:
        full_tree = ET.iterparse(filename)
        logger.info('Processing label: ' + filename)
    except IOError:
        raise IOError('Unable to locate or read label file: ' + filename)

    try:
        for _, el in full_tree:
            if '}' in el.tag:
                el.tag = el.tag.split('}', 1)[1]
        label_xml_root = full_tree.root
    except ExpatError:
        raise ExpatError('The requested label file does not appear contain valid XML: ' + filename)

    # Extract directory of label
    label_dir = path.dirname(filename)

    # Variable that stores byte data for each data file_xml referenced in the label
    byte_data = []

    # Loop over each File Area Observation and read-in all data files
    for i, file_xml in enumerate(label_xml_root.findall('.//File_Area_Observational')):

        data_filename = file_xml.find('.//file_name').text
        data_filepath = path.join(label_dir, data_filename)

        logger.info('Now reading data file_xml: ' + data_filename)

        # Read in the entire data file_xml in binary form into byte_data
        try:
            file_handler = open(data_filepath, 'rb')
            file_byte_data = file_handler.read()
            file_handler.close()

            byte_data.append(file_byte_data)
        except IOError:
            raise IOError("Unable to read data file '" + data_filepath + "' found in label")

    # Convert byte_data to tuple (in order to a pass a non-mutable variable as argument into read_table and read_array)
    byte_data = tuple(byte_data)

    # Variable that stores all returned data
    data = DataStructure()

    # Search for table types contained in label and retrieve data from them
    table_types_avail = ['Character', 'Binary']
    table_types_found = []

    for table_type in table_types_avail:

        if label_xml_root.find('.//Table_' + table_type) is not None:
            table_types_found.append(table_type)

    for table_type in table_types_found:

        single_type_data = pds4_read_table.read_table(label_xml_root, byte_data, table_type, use_numpy=use_numpy,
                                                      quiet=quiet, object_num=object_num, object_name=object_name,
                                                      object_lid=object_lid)

        for table_name, table_data in single_type_data.iteritems():
            data.set(table_name, table_data, single_type_data.meta_data(table_name))

    return data

if __name__ == '__main__':

    # Safe import of argparse
    try:
        import argparse
    except ImportError:
        import pds4_tools.argparse_26 as argparse

    # Create program arguments
    parser = argparse.ArgumentParser()
    parser.add_argument("filename", help="Filename, including full path, of the label")
    parser.add_argument("--use_numpy", help="Returned data will be a numpy array and use numpy data types",
                        action="store_true", default=False)
    parser.add_argument("--quiet", help="Suppresses all info/warnings", action="store_true", default=False)
    parser.add_argument("--object_num", help="Only reads the data object specified by zero-based order (integer)")
    parser.add_argument("--object_name", help="Only reads the data object specified by name")
    parser.add_argument("--object_lid", help="Only reads the data object specified by local identifier")

    args = parser.parse_args()

    pds4_read(args.filename, use_numpy=args.use_numpy, quiet=args.quiet, object_num=args.object_num,
              object_name=args.object_name, object_lid=args.object_lid)