from math import log

import pds4_read_table_ascii
import pds4_read_table_binary
from pds4_general_objects import DataStructure
from pds4_table_objects import TableStructure, TableDataObject, UniformlySampledField, is_group
from pds4_helpers import is_int
from pds4_logging import logger_init

# Safe import of numpy (not required)
try:
    import numpy as np
except ImportError:
    pass

# Initialize the logger
logger = logger_init()

#################################

# Creates the storage variable, including all stored data, for Uniformly_Sampled
def _make_uniformly_sampled(table_xml, use_numpy=False):
    uniformly_sampled_xml = table_xml.find('Uniformly_Sampled')
    uniformly_sampled_data = []
    uniformly_sampled_dict = {}

    if uniformly_sampled_xml is not None:

        uni_sampled = UniformlySampledField(uniformly_sampled_xml)

        if uni_sampled.scale == 'Linear':
            current_value = uni_sampled.first_value

            while current_value <= uni_sampled.last_value:
                uniformly_sampled_data.append(current_value)
                current_value += uni_sampled.interval

        elif uni_sampled.scale == 'Logarithmic':

            x = log(uni_sampled.first_value) / log(10)
            current_value = uni_sampled.first_value
            counter = 0

            while current_value < uni_sampled.last_value:
                current_value = 10 ** (uni_sampled.interval * counter + x)
                counter += 1

                uniformly_sampled_data.append(current_value)

        elif uni_sampled.scale == 'Exponential':

            x = 10. ** uni_sampled.first_value
            current_value = uni_sampled.first_value
            counter = 0

            while current_value < uni_sampled.last_value:
                current_value = log(uni_sampled.interval * counter + x) / log(10)
                counter += 1

                uniformly_sampled_data.append(current_value)

        # Return data values to int if this was intended
        if is_int(uni_sampled.interval) and is_int(uni_sampled.first_value) and is_int(uni_sampled.last_value):

            all_ints = True

            for value in uniformly_sampled_data:
                if not value.is_integer():
                    all_ints = False

            if all_ints:
                uniformly_sampled_data = [int(value) for value in uniformly_sampled_data]

        if use_numpy:
            uniformly_sampled_dict = {uni_sampled.name: np.asarray(uniformly_sampled_data)}
        else:
            uniformly_sampled_dict = {uni_sampled.name: tuple(uniformly_sampled_data)}

    return uniformly_sampled_dict


# Extracts and properly formats the data for a single table, as specified by table_object, from byte_data
def _extract_table_data(file_byte_data, table_object, table_structure, table_xml, use_numpy=False):

    # Variable that will store final version of all returned data from this table
    data = DataStructure()

    # Add Uniformly_Sampled to data if it exists
    uniformly_sampled = _make_uniformly_sampled(table_xml, use_numpy)
    if uniformly_sampled:
        data.set(uniformly_sampled.keys()[0], uniformly_sampled.values()[0], None)

    # This for loop creates the 'data' variable, which will store all columns and their values for the table that was
    # just read in. Begin by looping over each column
    for i in range(0, table_structure.num_items):

        # Skip groups; we need the i to be the index in table_structure below but do not need groups
        if is_group(table_structure[i]):
            continue

        field = table_structure[i]
        extracted_bytes = []

        # Extract the byte data for this specific column from byte_data
        for a in range(0, table_object.records):
            start_idx = table_object.offset + (field.location - 1) + (a * table_object.record_length)
            end_idx = start_idx + field.length

            extracted_bytes.append(file_byte_data[start_idx:end_idx])

        # Cast the byte data into appropriate data times
        if table_object.type == 'Table_Character':
            extracted_data = pds4_read_table_ascii.data_type_convert(field.data_type, extracted_bytes, use_numpy)

        elif table_object.type == 'Table_Binary':
            extracted_data = pds4_read_table_binary.data_type_convert(field.data_type, extracted_bytes, use_numpy)

        # Make scaling_factor and value_offset adjustments
        if field.scaling_factor or field.value_offset:

            for a, datum in enumerate(extracted_data):
                if field.scaling_factor:
                    extracted_data[a] *= field.scaling_factor

                if field.value_offset:
                    extracted_data[a] += field.value_offset

        # Retrieve indicies of all the parent groups of this column (needed to extract the group names for the 'data'
        # variable)
        parent_group_idxs = []

        for a in range(0, field.group_level):

            if parent_group_idxs:
                parent_idx = table_structure.get_parent_by_idx(parent_group_idxs[a - 1], return_idx=True)
            else:
                parent_idx = table_structure.get_parent_by_idx(i, return_idx=True)

            parent_group_idxs.append(parent_idx)

        # Find the appropriate place in the 'data' variable to store the newly extracted column by going deeper into
        # the 'data' variable (e.g. from data['GROUP_0'] to data['GROUP_0']['GROUP_0']) with each iteration of the loop
        data_ref = data

        for a in range(len(parent_group_idxs) - 1, -1, -1):
            group = table_structure[parent_group_idxs[a]]

            if group.name not in data_ref:
                data_ref.set(group.name, DataStructure(), group)

            data_ref = data_ref[group.name]

        # Insert the newly extracted column into the 'data' variable
        if not use_numpy:
            extracted_data = tuple(extracted_data)

        # Correctly append this field
        if field.name not in data_ref:
            data_ref.set(field.name, extracted_data, field)

        # Correctly append this field if this is a repetition of an existing field
        else:

            # Append to Field that is currently a nested tuple and thus already had a repetition prior to this one
            if hasattr(data_ref[field.name][0], '__len__') and (not isinstance(data_ref[field.name][0], str)):

                field_metadata = (data_ref.meta_data(field.name)) + (field,)
                if not use_numpy:
                    data_ref.set(field.name, (data_ref[field.name]) + (extracted_data,), field_metadata)

            # Append to Field that is currently tuple and thus has not already had repetition prior to this one
            else:

                field_metadata = (data_ref.meta_data(field.name),) + (field,)
                if not use_numpy:
                    data_ref.set(field.name, (data_ref[field.name],) + (extracted_data,), field_metadata)

            if use_numpy:
                data_ref.set(field.name, np.vstack((data_ref[field.name], extracted_data)), field_metadata)

    return data


def read_table(label_xml_root, file_byte_data, table_type, use_numpy=False, quiet=False, object_num=None,
               object_name=None, object_lid=None):

    # Set logger to only emit error messages if requested
    if quiet:
        logger.quiet()

    # Storage variable for the data to be read
    data = DataStructure()

    # Loop over each File Area Observation
    for i, file_xml in enumerate(label_xml_root.findall('.//File_Area_Observational')):

        data_filename = file_xml.find('.//file_name').text

        # Loop over each Table inside the File Area
        for j, table in enumerate(label_xml_root.findall('.//Table_' + table_type)):

            # Determine the name of table
            table_name_xml = table.find('name')

            if table_name_xml is not None:
                table_name = table_name_xml.text

                if table_name in data:
                    logger.warn('found duplicate data object name ' + table_name + ', object renamed')
                    table_name += ' ' + str(j)

            else:
                table_name = 'TABLE_' + str(j)

            # Create the data object for this table
            table_object = TableDataObject(table_name, 'Table_' + table_type, data_filename, table)

            # Determine if this table should be saved
            if object_num is not None and object_num != j:
                continue

            if object_name is not None and table_object.name != object_name:
                continue

            if object_lid is not None and table_object.local_identifier != object_lid:
                continue

            logger.info('Now processing a Table_{0} object with name: {1}'.format(table_type, table_name))

            # Stores data about the table's groups and fields
            record_character = table.find('Record_' + table_type)
            table_structure = TableStructure()
            table_structure.load(table_object.type, record_character)

            # Read the table from file into a variable
            single_table_data = _extract_table_data(file_byte_data[i], table_object, table_structure, table, use_numpy)

            # Save the read-in table and its meta data
            data.set(table_name, single_table_data, table_object)

    if len(data) == 0:
        raise ValueError('No viable Table objects found')

    return data
