import copy
import re
from collections import Sequence

from pds4_general_objects import DataObject
from pds4_helpers import cast_int_or_float
from pds4_logging import logger_init

# Initialize the logger
logger = logger_init()

#################################


class TableStructure(Sequence):
    """ Stores meta data about a single table's groups and fields

    Once .load() is used, this object will read the XML Record for a single table and create a representation
    of the table, ordered by byte location, by adding the appropriate Field and Group objects.
    """

    def __init__(self):

        self._struct = []
        self._num_items = 0

        self._table_type = None
        self._xml_record_character = None

        super(TableStructure, self).__init__()

    def __getitem__(self, key):
        return self._struct[key]

    def __len__(self):
        return self._num_items

    # Fully initializes the TableStructure with all appropriate Field and Group objects
    def load(self, object_type, xml_record_character):

        self._table_type = object_type.split('_')[1]
        self._xml_record_character = xml_record_character

        # Add all fields and groups to the TableStructure
        self._add_child_fields(xml_record_character)
        self._add_child_groups(xml_record_character)

        # Create a pre-cloned copy, used for validation
        pre_cloned_table_struct = copy.deepcopy(self)
        pre_cloned_table_struct.create_group_names()
        pre_cloned_table_struct.create_full_names()

        # Clone repeating group fields and adjust locations to their absolute values instead of the default
        # relative values
        self._clone_all_groups()
        self._adjust_absolute_location()

        # Into TableStructure insert group names for groups, and insert full names for all fields and groups
        self.create_group_names()
        self.create_full_names()

        # Perform basic sanity and basic validation checking on the table
        self._validate_table(pre_cloned_table_struct)
        self._validate_table_sbn(pre_cloned_table_struct)

    @property
    def struct(self):
        return self._struct

    @property
    def num_items(self):
        return self._num_items

    # Returns a new TableStructure that contains the children of the idx element. Returns only the immediate children
    # if direct_only is set to True
    def get_children_by_idx(self, parent_idx, direct_only=False):

        child_table_structure = TableStructure()
        parent_element_group_lvl = self._struct[parent_idx].group_level + 1

        for i in range(parent_idx + 1, self._num_items):
            cur_element = self._struct[i]

            if cur_element.group_level >= parent_element_group_lvl:

                if cur_element.group_level == parent_element_group_lvl or not direct_only:
                    child_table_structure._append(cur_element)

            else:
                break

        return child_table_structure

    # Returns the parent group of the child given by child_idx. Returns the index of the parent group if  return_idx
    # is set to True.
    def get_parent_by_idx(self, child_idx, return_idx=False):
        child_element = self._struct[child_idx]
        child_group_level = child_element.group_level

        for i in range(child_idx - 1, -1, -1):
            cur_element = self._struct[i]

            if (cur_element.group_level + 1) == child_group_level and is_group(cur_element):

                if return_idx:
                    return i
                else:
                    return cur_element

        return -1

    # Recursively creates names for groups if they do not exist in the label; created group names are only unique in
    # their current group_level and for the same parent
    def create_group_names(self, group_level=0):

        group_indicies = [i for i, item in enumerate(self._struct)
                          if is_group(item)
                          if item.group_level == group_level]

        if not group_indicies:
            return

        prev_parent = -2
        counter = 0

        for group_idx in group_indicies:

            group = self._struct[group_idx]
            parent_idx = self.get_parent_by_idx(group_idx, return_idx=True)

            if parent_idx == prev_parent:
                counter += 1
            else:
                counter = 0

            if group.name is None:
                group.name = 'GROUP_' + str(counter)
                group.full_name = group.name

            prev_parent = parent_idx

        self.create_group_names(group_level + 1)

    # Creates the full names (ie, name preceded by all parent group names) for all fields and groups
    def create_full_names(self):

        for i, item in enumerate(self._struct):

            full_name = item.name
            idx = i

            while True:
                parent_idx = self.get_parent_by_idx(idx, return_idx=True)

                if parent_idx != -1:
                    idx = parent_idx
                    full_name = self._struct[parent_idx].name + ': ' + full_name
                else:
                    item.full_name = full_name
                    break

    def _insert(self, key, field_or_group):
        self._struct.insert(key, field_or_group)
        self._num_items += 1

    def _append(self, field_or_group):
        self._struct.append(field_or_group)
        self._num_items += 1

    # Adds <Field_*> to TableStructure which are direct children of xml_parent
    def _add_child_fields(self, xml_parent, group_level=0):

        for field_character in xml_parent.findall('Field_' + self._table_type):
            appended_field = Field(field_character)
            self._append(appended_field)

            appended_field.group_level = group_level

    # Adds all <Group_Field_*>, and all sub-<Field_*> and sub-<Group_Field_*>, to TableStructure which are direct
    # children of xml_parent
    def _add_child_groups(self, xml_parent, group_level=0):

        for group_character in xml_parent.findall('Group_Field_' + self._table_type):

            appended_group = Group(group_character)
            self._append(appended_group)

            appended_group.group_level = group_level
            self._add_child_fields(group_character, group_level + 1)

            if group_character.findall('Group_Field_' + self._table_type):
                self._add_child_groups(group_character, group_level + 1)

    # Clones/repeats all <Group_Field_*> in TableStructure according to the 'repetition' field
    def _clone_all_groups(self):
        index_adjust = 0
        group_indicies = [i for i, item in enumerate(self._struct)
                          if is_group(item)
                          if not item.cloned]

        if not group_indicies:
            return

        for group_idx in group_indicies:
            start_length = self._num_items
            self._clone_group(group_idx + index_adjust)
            index_adjust += self._num_items - start_length

        self._clone_all_groups()

    # Clones a single <Group_Field_*>, adjusting location to account for the repetition
    def _clone_group(self, idx):
        group = self._struct[idx]

        cloned_group = copy.deepcopy(self.get_children_by_idx(idx))
        current_idx = idx + cloned_group.num_items

        for i in range(1, group.repetitions):

            # Insert each child of the group being cloned individually
            for item in copy.deepcopy(cloned_group):

                current_idx += 1
                self._insert(current_idx, item)
                parent_idx = self.get_parent_by_idx(current_idx, return_idx=True)

                # If element is a direct child of the group being cloned then adjust its location
                if parent_idx == idx:
                    group_length = self._struct[parent_idx].length
                    group_repetitions = self._struct[parent_idx].repetitions

                    item.location += group_length / group_repetitions * i

        group.cloned = True

    # Adjusts group_location and field_location in TableStructure to be the absolute locations from the beginning of
    # the data file instead of the relative location used by PDS4 schemas inside <Group_Field_*>
    def _adjust_absolute_location(self):
        group_indicies = [i for i, item in enumerate(self._struct)
                          if is_group(item)
                          if item.group_level == 0]

        for group_idx in group_indicies:
            child_elements = self.get_children_by_idx(group_idx).struct

            for i, element in enumerate(child_elements):

                parent_idx = self.get_parent_by_idx(group_idx + i + 1, return_idx=True)

                if parent_idx != -1:
                    element.location += self._struct[parent_idx].location - 1

    # Performs basic sanity checks on the table and basic verification of some PDS4 standards not testing by schema
    def _validate_table(self, pre_cloned_table_struct):

        # Check that number of fields and groups claimed matches number of <Field_*> and <Group_Field_*>
        num_rec_fields = int(self._xml_record_character.find('fields').text)
        num_rec_groups = int(self._xml_record_character.find('groups').text)

        groups = [item for item in pre_cloned_table_struct if is_group(item) and item.group_level == 0]
        fields = [item for item in pre_cloned_table_struct if not is_group(item) and item.group_level == 0]

        fields_mismatch_warn = '<fields> value does not match number of <Field_{0}> in'.format(self._table_type)
        groups_mismatch_warn = '<groups> value does not match number of <Group_Field_{0}> in'.format(self._table_type)

        if len(fields) != num_rec_fields:
            logger.warning('{0} Record_{1}.'.format(fields_mismatch_warn, self._table_type))

        if len(groups) != num_rec_groups:
            logger.warning('{0} Record_{1}.'.format(groups_mismatch_warn, self._table_type))

        for i, item in enumerate(pre_cloned_table_struct):

            full_location = item.full_name.replace(': ', ' -> ')
            full_location_warn = item.name + " (full location: '" + full_location + "')"

            if is_group(item):

                # Check number of fields and groups matching in nested groups
                children = pre_cloned_table_struct.get_children_by_idx(i, direct_only=True)

                groups = [child for child in children if is_group(child)]
                fields = [child for child in children if not is_group(child)]

                if len(fields) != item.fields:
                    logger.warning(fields_mismatch_warn + 'group ' + full_location_warn)

                if len(groups) != item.groups:
                    logger.warning(groups_mismatch_warn + 'group ' + full_location_warn)

                # Check that group_length is evenly divisible by group_repetitions
                if item.length % item.repetitions != 0:
                    raise Exception("Group length '{0}' must be evenly divisible by the number of repetitions '{1}' "
                                    "for group {2}".format(item.length, item.repetitions, full_location_warn))

            elif item.format is not None:

                # Ensure field_format requirements
                valid_field_format = re.compile('%([\+,-])?([0-9]+)(\.([0-9]+))?([doxfes])')

                if not valid_field_format.match(item.format):
                    logger.warning("field_format '{0}' does not conform to standards for field {1} "
                                   .format(item.format, full_location_warn))

                else:

                    is_numeric = {'ASCII_Real': True,
                                  'ASCII_Integer': True,
                                  'ASCII_NonNegative_Integer': True,
                                  'ASCII_Numeric_Base2': True,
                                  'ASCII_Numeric_Base8': True,
                                  'ASCII_Numeric_Base16': True}.get(item.data_type, False)

                    sub_exp = valid_field_format.search(item.format).groups()

                    plus_minus = sub_exp[0]
                    width = sub_exp[1]
                    precision = sub_exp[3]
                    specifier = sub_exp[4]

                    if plus_minus is not None:

                        if plus_minus == '+' and not is_numeric:
                            logger.warning("field_format '{0}' may not have '+' for a numeric data_type '{1}' for "
                                           "field {2}".format(item.format, item.data_type, full_location_warn))

                        elif plus_minus == '-' and is_numeric:
                            logger.warning("field_format '{0}' may not have '-' for a non-numeric data_type '{1}' for "
                                           "field {2}".format(item.format, item.data_type, full_location_warn))

                    if int(width) != item.length and self._table_type == 'Character':
                        logger.warning("in Table_Characters field_format '{0}' width must be equal to field_length "
                                       "'{1}' for field {2}".format(item.format, item.length, full_location_warn))

                    if precision is not None and specifier == 's' and int(width) != int(precision):
                        logger.warning("field_format '{0}' precision must be equal to width '{1}' for a string format "
                                       "for field {2}".format(item.format, width, full_location_warn))

    # Additional sanity checks on TableStructure, looking for invalid content according to PDS-SBN standards but
    # valid PDS4
    def _validate_table_sbn(self, pre_cloned_table_struct):

        groups = [item for item in pre_cloned_table_struct if is_group(item)]

        if groups:
            logger.warning('-SBN-: Group fields are used by this table')

        for i, item in enumerate(pre_cloned_table_struct):

            full_location = item.full_name.replace(': ', ' -> ')
            full_location_warn = item.name + " (full location: '" + full_location + "')"

            if is_group(item):

                if item.repetitions < 2:
                    logger.warning("-SBN-: A group's repetitions should be at least 2 for Group " + full_location_warn)

            else:

                # Check that expected elements for each Field exist and conform to SBN standards

                if not groups and item.number is None:
                    logger.warning('-SBN-: a field_number is expected for all Fields in flat tables, but one was not '
                                   'found for Field ' + full_location_warn)

                if item.format is None:
                    logger.warning('-SBN-: a field_format is required for all Fields, but was not found for Field '
                                   + full_location_warn)

                if item.description is None:
                    logger.warning('-SBN-: description is expected for all Fields but was not found for Field '
                                   + full_location_warn)

                if item.unit is not None and (item.unit == 'N/A' or item.unit.strip() == ''):
                    logger.warning('-SBN-: blank or N/A should not be used for units but was found for Field '
                                   + full_location_warn)

        # Check for overlapping Field locations
        fields = [item for item in self._struct if not is_group(item)]

        if len(fields) > 1000:
            logger.warning('-SBN-: Large number of Fields, field_location overlap checking skipped.')
            return

        for i, item in enumerate(fields):

            for j, inner_item in enumerate(fields):

                if i == j:
                    continue

                item_end_location = item.location + item.length - 1

                if item.location <= inner_item.location <= item_end_location:

                    full_location = item.full_name.replace(': ', ' -> ')
                    full_location_warn = item.name + " (full location: '" + full_location + "')"

                    full_location_inner = inner_item.full_name.replace(': ', ' -> ')
                    full_location_inner_warn = inner_item.name + " (full location: '" + full_location_inner + "')"

                    logger.warning('-SBN-: Overlapping field_locations found for Field {0} and Field {1}'
                                   .format(full_location_warn, full_location_inner_warn))


class TableDataObject(DataObject):
    """ Stores meta data about a single Table Object (i.e, Table_Character, Table_Binary or Table_Delimited) """

    def __init__(self, name, object_type, parent_filename, table_xml):

        super(TableDataObject, self).__init__(name, object_type, parent_filename)

        self.records = int(table_xml.find('records').text)
        self.offset = int(table_xml.find('offset').text)

        self.local_identifier = None
        self.record_length = None
        self.field_delimiter = None
        self._set_optional_fields(object_type, table_xml)

    def _set_optional_fields(self, object_type, table_xml):

        if table_xml.find('local_identifier') is not None:
            self.local_identifier = table_xml.find('local_identifier').text

        table_type = object_type.split('_')[1]
        record = table_xml.find('Record_' + table_type)

        if record.find('record_length') is not None:
            self.record_length = int(record.find('record_length').text)

        if record.find('field_delimiter') is not None:
            self.record_length = record.find('field_delimiter').text


class Field(object):
    """ Stores meta data about a single <Field_*> """

    def __init__(self, xml_field):

        self.name = xml_field.find('name').text
        self.full_name = self.name
        self.location = int(xml_field.find('field_location').text)
        self.length = int(xml_field.find('field_length').text)
        self.data_type = xml_field.find('data_type').text
        self.group_level = 0

        self.number = None
        self.format = None
        self.scaling_factor = None
        self.value_offset = None
        self.unit = None
        self.description = None
        self._set_optional_attributes(xml_field)

    def _set_optional_attributes(self, xml_field):

        if xml_field.find('field_number') is not None:
            self.number = int(xml_field.find('field_number').text)

        if xml_field.find('field_format') is not None:
            self.format = xml_field.find('field_format').text

        if xml_field.find('scaling_factor') is not None:
            self.scaling_factor = cast_int_or_float(xml_field.find('scaling_factor').text)

        if xml_field.find('value_offset') is not None:
            self.value_offset = cast_int_or_float(xml_field.find('value_offset').text)

        if xml_field.find('unit') is not None:
            self.unit = xml_field.find('unit').text

        if xml_field.find('description') is not None:
            self.description = xml_field.find('description').text


class UniformlySampledField(object):
    """ Stores meta data about a single <Uniformly_Sampled> """

    def __init__(self, xml_uniformly_sampled):

        self.name = xml_uniformly_sampled.find('sampling_parameter_name').text
        self.full_name = self.name
        self.interval = cast_int_or_float(xml_uniformly_sampled.find('sampling_parameter_interval').text)
        self.unit = xml_uniformly_sampled.find('sampling_parameter_unit').text
        self.first_value = cast_int_or_float(xml_uniformly_sampled.find('first_sampling_parameter_value').text)
        self.last_value = cast_int_or_float(xml_uniformly_sampled.find('last_sampling_parameter_value').text)

        self.scale = 'Linear'
        self._set_optional_attributes(xml_uniformly_sampled)

    def _set_optional_attributes(self, xml_uniformly_sampled):

        if xml_uniformly_sampled.find('sampling_parameter_scale') is not None:
            self.scale = xml_uniformly_sampled.find('sampling_parameter_scale').text


class Group(object):
    """ Stores meta data about a single <Group_Field_*> """

    def __init__(self, xml_group_field):

        self.location = int(xml_group_field.find('group_location').text)
        self.length = int(xml_group_field.find('group_length').text)
        self.repetitions = int(xml_group_field.find('repetitions').text)
        self.fields = int(xml_group_field.find('fields').text)
        self.groups = int(xml_group_field.find('groups').text)

        self.cloned = False
        self.group_level = 0

        self.name = None
        self.full_name = None
        self._set_optional_attributes(xml_group_field)

    def _set_optional_attributes(self, xml_group_field):

        if xml_group_field.find('name') is not None:
            self.name = xml_group_field.find('name').text
            self.full_name = self.name

# Returns true if table_element is a Group object (i.e. Group_Field_*), and false if it is not
def is_group(table_element):
    if isinstance(table_element, Group):
        return True

    return False
